#include <vector>
#include <stdlib.h> /* rand() */
#include <time.h>

#define BISQBOT_INTERNAL
#include "bisqbot.hh"
static const unsigned MAX_FRAMES = 240;

extern "C" unsigned char RAM[0x800];

static unsigned NextChange=0;
static unsigned NextJump  =0;

static class BisqBotState: public BisqBotStateBase
{
public:
    virtual ~BisqBotState() { }
    
    BisqBotState()
    {
        Reset();
    }
    void Reset()
    {
        PrevMarioX = 0;
        NFrames    = 0;
        MarioMovement = 0;
        Pit = 0;
        Entry.Clear();
        Damage.Clear();
        
        JumpRemain = 0;
    }
    
    void operator= (const BisqBotState& b)
    {
        PrevMarioX    = b.PrevMarioX;
        NFrames       = b.NFrames;
        MarioMovement = b.MarioMovement;
        Pit = b.Pit;
        Entry = b.Entry;
        Damage = b.Damage;
        JumpRemain = b.JumpRemain;
    }

public:
    unsigned PrevMarioX, NFrames;
    int MarioMovement;
    int Pit;
    BisqBotVar Entry, Damage;

    unsigned JumpRemain;
} state;

const BisqBotStateBase* BisqBotNewSaveState()
{
    return new BisqBotState(state);
}

void BisqBotLoadState(const BisqBotStateBase* b)
{
    state = *(BisqBotState*)b;
}

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
    /* GAME MONITORING */
    /* A, X, Y, PC correspond to the values of the CPU registers. */
    if(PC == 0x8082)
    {
        unsigned MarioXPos = RAM[0x86];
        unsigned MarioYPos = RAM[0xCE];
        unsigned ScrollBegin = RAM[0x73F];
        
        if(!state.Pit && MarioYPos >= 0xF0 && MarioYPos <= 0xFF)
        {
            state.Pit = -1; // ignore, it's screen top
        }
        else if(state.Pit < 0 && MarioYPos < 0x60)
            state.Pit = 0;
        else if(!state.Pit && MarioYPos > 0xC0 && MarioYPos < 0xFF)
        {
            state.Pit=1;
            fprintf(stderr, "Pit (y=%u)\n", MarioYPos);
        }
        
        unsigned MarioScreenX = (unsigned char)(MarioXPos-ScrollBegin+256);
        
        unsigned MarioX = MarioXPos;
        if(state.NFrames++ > 0)
        {
            //fprintf(stderr, "MarioX = %u\n", MarioX);
            state.MarioMovement += (signed char)(MarioX - state.PrevMarioX);
        }
        state.PrevMarioX = MarioX;
    }
    
    if(PC == 0xB04C) // magic_jT9+2
    {
        //fprintf(stderr, "JT9 with A=%02X\n", A);
        switch(A)
        {
            case 1: // auto climb
            case 2: // enter horiz pipe
            case 3: // enter vert pipe
            case 4: // slide flagpole down
            case 5: // victory 
            case 9: // eat mushroom
            case 12: // eat flower
                state.Entry.Set(A);
                break;
            case 10:// eat damage
            case 11:// eat death
                state.Damage.Set(A);
                fprintf(stderr, "Pit (a=%u)\n", A);
                state.Pit=1;
                break;
            case 0: // regen level
            case 7: // start level
                break;
            case 8: // normal level activity
                break;
        }
    }
    if(PC == 0xCFEC)
    {
        state.Entry.Set(15); // Killed koopa
    }
}

void BisqBotReset()
{
    state.Reset();
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    if(state.Entry.Count()) return 0;
    
    //if(FrameNo >= 4) return K_A + K_B + K_R;
    
    static const unsigned char InputTable[] =
    {
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,
        K_R,
        K_L,
        K_R | K_L,
        0
    };
    static unsigned LastInput = 0;
    static unsigned A_State   = 0;
    static unsigned B_State   = K_B;
    unsigned char Input = 0x00;
    
    if(!NextChange) 
    {
        LastInput = InputTable[rand()%sizeof(InputTable)];
        NextChange = rand()%3;
    }
    else --NextChange;
    
    /*B_State = K_B;*/
    /**/
    if(B_State)
        { if(rand()%100 < 1) B_State=0; }
    else
        { if(rand()%100 < 95) B_State=K_B; }
    /**/

    /**/
    if(!(rand()%20))
    {
        Input |= K_D;
        if((rand()%5)) Input &= ~(K_R | K_L);
    }
    /**/
    
    if(!state.JumpRemain)
    {
        if(!NextJump)
        {
            state.JumpRemain = rand()%(1+rand()%60);
            NextJump   = rand()%30 + 1;
        } else --NextJump;
    }
    
    Input |= LastInput;
    if(state.JumpRemain) Input |= K_A;
    Input |= B_State;
    
    return Input;
}

unsigned long long BisqBotEvaluate(unsigned FrameNo)
{
    //fprintf(stderr, "Joypad=%02X\n", RAM[0x6FC]);
    /*
    fprintf(stderr, "Evaluate %u, movement=%d, %u frames, nextchange=%u\n",
        FrameNo,
        state.MarioMovement,
        state.NFrames,
        NextChange);
    */
    
    if(state.Pit > 0) return 0; // dead
    
    unsigned score1 = 0;
    for(unsigned a=0; a<6; ++a)
        score1 = score1*10 + RAM[0x7DD+a];
    //fprintf(stderr, "score1: %u\n", score1);
    
    unsigned movescore = state.MarioMovement + 512;
    unsigned goalscore = state.Entry.Count() * 2048;
    return (movescore + goalscore) * 1000000ULL + score1;
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo)
{
    if(state.Pit > 0) return BisqBotRestart;
    
    if(FrameNo > MAX_FRAMES) return BisqBotRestart;
    
    //fprintf(stderr, "%u: %d\n", FrameNo, MarioMovement);
    return BisqBotContinue;
}
