#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sstream>
#include <vector>

#include "types.h"
extern "C" {
#include "state.h"
}
#include "fceu.h"

#define BISQBOT_INTERNAL
#include "bisqbot.hh"

#include "coroutine.h"

int BotFrontDisableVideo=0;

#define PART2_LENGTH  60
#define PART1_LENGTH  20
#define MAX_REFINE_PART1 200
#define MAX_REFINE_PART2 15

struct SaveState
{
    std::vector<unsigned char> Data;
    const BisqBotStateBase* Bot;
    
    SaveState() : Bot(0) { }
    ~SaveState() { delete Bot; }

    void Create()
    {
        delete Bot;
        
        FILE*fp = tmpfile();
        FCEUSS_SaveFP(fp);
        rewind(fp);
        
        fseek(fp,0,SEEK_END);
        long pos = ftell(fp);
        rewind(fp);
        
        SaveState result;
        Data.resize(pos);
        fread(&Data[0], 1, pos, fp);
        fclose(fp);
        
        Bot = BisqBotNewSaveState();
    }
    void Load()
    {
        FILE*fp = tmpfile();
        fwrite(&Data[0], 1, Data.size(), fp);
        rewind(fp);
        FCEUSS_LoadFP(fp, 0);
        fclose(fp);
        
        BisqBotLoadState(Bot);
    }
private:
    SaveState(const SaveState& b);
    void operator=(const SaveState& b);
};

static unsigned char CurInput;
static bool Active=false;

static void DisplaySeq(const std::vector<unsigned char>& Seq)
{
    std::stringstream msg;
    for(unsigned a=0; a<Seq.size(); )
    {
        unsigned c=1, ch=Seq[a];
        while(a+c < Seq.size() && Seq[a+c] == ch) ++c;
        msg << c; msg << '*';
        
        unsigned b=0;
        if(ch & K_A) { if(b++) msg << ("_"); msg << ("A"); }
        if(ch & K_B) { if(b++) msg << ("_"); msg << ("B"); }
        if(ch & K_SE) { if(b++) msg << ("_"); msg << ("SE"); }
        if(ch & K_ST) { if(b++) msg << ("_"); msg << ("ST"); }
        if(ch & K_U) { if(b++) msg << ("_"); msg << ("U"); }
        if(ch & K_D) { if(b++) msg << ("_"); msg << ("D"); }
        if(ch & K_L) { if(b++) msg << ("_"); msg << ("L"); }
        if(ch & K_R) { if(b++) msg << ("_"); msg << ("R"); }
        if(!b) msg << ("00");
        msg << (", ");
        a += c;
    }
    BotFrontMsg("%s", msg.str().c_str());
}

void BotFrontMsg(const char*fmt, ...)
{
    FILE*fp;
    
    va_list ap;
    va_start(ap,fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    
    fp = fopen("rockman.log", "at");
    if(!fp) fp = fopen("rockman.log", "wt");

    va_start(ap,fmt);
    vfprintf(fp, fmt, ap);
    va_end(ap);
    
    fclose(fp);     
}

void BotFrontStart()
{
    fprintf(stderr,"BotFrontStart\n");
    BotFrontDisableVideo=1;
    BotFrontMsg("--------------- New run  ------------------ \n");
    srand(time(NULL));
    Active=true;
}

void BotFrontEnd()
{
    fprintf(stderr,"BotFrontEnd\n");
    BisqBotReset();
    BotFrontDisableVideo=0;
    Active=false;
} 

void BotFrontToggle()
{
    if(!Active)
        BotFrontStart();
    else
        BotFrontEnd();
}

unsigned BotFrontGetKeymask()
{
    if(Active)
    {
        //fprintf(stderr, "CurInput: %02X\n", CurInput);
        return CurInput;
    }
    //fprintf(stderr, "Nul keymask\n");
    return 0;
}

/*
    FCEU main loop:
        1. Emulate frame     
        2. Read keyboard input (BotFrontIter)
        3. Update joypad input (BotFrontGetKeymask)

    It is actually possible for keyboard input to be
    read multiple times per frame... Currently we hope
    this don't happen.
*/

void BotFrontIter()
{
    scrBegin;
    while(!Active)
    {
        BisqBotEvaluate(0);
        BisqBotReset();
        scrReturnV;
    }

    /* Create a sequence */
    static std::vector<unsigned char> BestSeq;
    static BisqBotInstruction verdict;
    static unsigned long long BestScore;
    static unsigned reattempt;
    static unsigned frameno;
    
    if(!FCEUI_EmulationPaused()) FCEUI_ToggleEmulationPause();
    

Repeat:
    BisqBotReset();
    static SaveState anchor;
    anchor.Create();
    
    #define RunFrameWith(k) \
        CurInput = (k); \
        FCEUI_FrameAdvance(); \
        scrReturnV;
    
    BestSeq.clear();
    BestScore = 0;
    
    // Attempt a few times
    for(reattempt=0; reattempt<MAX_REFINE_PART1; ++reattempt)
    {
        static std::vector<unsigned char> Seq;
        
        Seq.clear();
        anchor.Load();
        BisqBotReset();
        
        // Create the run
        for(frameno=0; frameno<PART1_LENGTH; ++frameno)
        {
            //fprintf(stderr, "attempt %u frameno %u\n", reattempt, frameno);
        
            verdict = BisqBotTick(frameno);
            if(verdict == BisqBotRestart)
            {
                goto TryOver;
            }
            else if(verdict == BisqBotAccept)
            {
                BotFrontMsg("\n***  Bingo in %u frames:\n", frameno);
                DisplaySeq(Seq);
                BotFrontMsg("\n\n");
                goto End;
            }
            else if(verdict == BisqBotContinue)
            {
                static unsigned char k;
                k = BisqBotInput(frameno);
                Seq.push_back(k);
                RunFrameWith(k);
            }
        }
        
        /* A candidate has been created. Find out its potential! */
        
        static SaveState subanchor;
        subanchor.Create();
        
        static unsigned long long best_subscore;
        static unsigned subattempt;
        static unsigned subframeno;
        best_subscore = 0;
        for(subattempt=0; subattempt<MAX_REFINE_PART2; ++subattempt)
        {
            Seq.resize(frameno);
            subanchor.Load();
            
            for(subframeno=0; subframeno<PART2_LENGTH; ++subframeno)
            {
                verdict = BisqBotTick(subframeno+frameno);
                if(verdict == BisqBotRestart) goto SubTryOver;
                if(verdict == BisqBotAccept)
                {
                    BotFrontMsg("\n***  Bingo in %u frames:\n", frameno+subframeno);
                    DisplaySeq(Seq);
                    BotFrontMsg("\n\n");
                    goto End;
                }
                else if(verdict == BisqBotContinue)
                {
                    static unsigned char k;
                    k = BisqBotInput(subframeno+frameno);
                    Seq.push_back(k);
                    RunFrameWith(k);
                }
            }
            
            if(true)
            {
                unsigned long long score = BisqBotEvaluate(subframeno+frameno);
                if(score > best_subscore)
                    best_subscore = score;
            }
            //fprintf(stderr, "(%u,%u)%llu...\n", frameno,subframeno,score);
        SubTryOver: ;
        }
        
        Seq.resize(frameno);
        
        // Score this run
        if(best_subscore > BestScore)
        {
            BotFrontMsg("Try %d: score %llu (%u frames)\n",
                reattempt, best_subscore, frameno);
            DisplaySeq(Seq);
            BotFrontMsg("\n");
            BestScore = best_subscore;
            BestSeq = Seq;

            /* It's a new record! Therefore, do more refining! */
            reattempt = 0;
        }
        else if(best_subscore == BestScore)
        {
            /* Bleh desu, it's the same again */
            BotFrontMsg("[%u] ", reattempt);
            DisplaySeq(Seq);
            BotFrontMsg("\n");
        }
        
        #if 0
        if(BestScore < 640000000ULL)
        {
            // Don't accept this input!
            --reattempt;
        }
        #endif
     
    TryOver: ;
    }
    
    anchor.Load();
    
    // Repeat it
    for(frameno=0; frameno<PART1_LENGTH && frameno<BestSeq.size(); ++frameno)
    {
        RunFrameWith(BestSeq[frameno]);
    }

    // Go and repeat the thing.
    goto Repeat;
    
    #undef RunFrameWith

End: ;
    BotFrontEnd();

    scrFinishV;
}
